
//
// For a least-squares fit of a straight line to a number of data
// points (x[k],y[k]), calculates slope a, intercept b, std
// errors in a and b (with Student t-factors incorporated),
// chi-square (where independent estimates of sigmas are
// supplied), and the correlation coefficient r.
//
// To reduce susceptibility to roundoff errors in the
// calculations, we follow the advice from Numerical Recipes,
// which involves the introduction of the variable z
//



#include "fitline.h"

// x is first data column, contains x-values
// y is second data column, contains y-values
// s is third column, contains sigmas (either indep or relative)


void calculate(dvec& x, dvec& y, dvec& s, opts& o, result& r)
{
  uns n1 = x.size();                                     // no. of x-values
  uns n2 = y.size();                                     // no. of y-values
  assert(n1 == n2);                     // ensure x and y are of equal size
  uns npts = n1;                                // total no. of data points
  uns nu = npts - 2;               // degs of freedom, equal to npts-2 here
  dvec z(n1);
  dvec w(n1);                                           // the weights w[k]
  uns k;
  dbl sw;                                                    // sum of w[k]
  dbl sx;                                               // sum of w[k]*x[k]
  dbl sy;                                               // sum of w[k]*y[k]
  dbl sxx;                                         // sum of w[k]*x[k]*x[k]
  dbl syy;                                         // sum of w[k]*y[k]*y[k]
  dbl sxy;                                         // sum of w[k]*x[k]*y[k]
  dbl sxosw;                                           // sx/sw, used below
  dbl szz;                 // sum of z[k]*z[k] = sum of w[k]*(x[k]-sxosw)^2
  dbl szy;                                 // sum of (w[k])^(0.5)*z[k]*y[k]
//dbl delta;                                  // determinant of coefficents
  dbl a;                                                           // slope
  dbl b;                                                       // intercept
  dbl phi;
  dbl corr;                                      // correlation coefficient
  dbl siga;                                               // std error in a
  dbl sigb;                                               // std error in b
  dbl chisq;                                                  // chi-square
  dbl alpha;                            // probability of larger chi-square
  cdbl conf = 0.68269;                                  // confidence level
  dbl t;                                                // student t-factor

  sw = sx = sy = sxx = syy = sxy = szz = szy = phi = 0.0;

  for (k = 0 ; k < n1 ; k++) {           // 1st pass: calculate needed sums
    if (r.ssize == 0) {                          // if sigmas not supplied,
      w[k] = 1.0;         // all weights are assumed equal; sw will be npts
    }
    else {                                               // sigmas supplied
      w[k] = 1.0/(s[k]*s[k]);           // sigmas supplied, so this is w[k]
    }
    sw += w[k];                                          // add up the w[k]
    sx += w[k]*x[k];                            // add up the weighted x[k]
    sy += w[k]*y[k];                            // add up the weighted y[k]
    sxx += w[k]*x[k]*x[k];                 // add up the weighted x[k]*x[k]
    syy += w[k]*y[k]*y[k];                 // add up the weighted y[k]*y[k]
    sxy += w[k]*x[k]*y[k];                 // add up the weighted x[k]*y[k]
  }
  sxosw = sx/sw;                                   // a convenient variable
  for (k = 0 ; k < n1 ; k++) {      // 2nd pass: add up z-related variables
    z[k] = sqrt(w[k])*(x[k]-sxosw);
    szz += z[k]*z[k];
    szy += sqrt(w[k])*z[k]*y[k];
  }
  t = student(nu, conf);                                // student t-factor
//delta = sw*sxx - sx*sx;  // used for testing
  a = szy/szz;
  b = (sy-sx*a)/sw;
  siga = t*sqrt(1.0/szz);
  sigb = t*sqrt((1.0+sx*sx/(sw*szz))/sw);
  for (k = 0 ; k < n1 ; k++) {                   // 3rd pass: determine phi
    phi += w[k]*(y[k]-a*x[k]-b)*(y[k]-a*x[k]-b);
  }
  corr = (sw*sxy-sx*sy)/(sqrt((sw*sxx-sx*sx)*(sw*syy-sy*sy)));
                           // corr is r, the linear correlation coefficient
//dbl phi_nr = (1.0 - corr*corr)*(syy - sy*sy/sw);   //interesting relation

  if (r.ssize != 0 && o.sigmas) {        // independent sigmas are supplied
    chisq = phi;                                      // chisq is available
    alpha = gammq(0.5*nu, 0.5*chisq);        // probability of larger chisq
  }
  else {        // no o.sigmas, and/or s is empty, no valid chisq available
    siga *= sqrt(phi/nu);
    sigb *= sqrt(phi/nu);
  }
  // put results in r:
  r.npts  = npts;
  r.a     = a;                                                     // slope
  r.b     = b;                                                 // intercept
  r.siga  = siga;                                     // std error of slope
  r.sigb  = sigb;                                 // std error of intercept
  r.corr  = corr;                                // correlation coefficient
  r.conf  = 100*conf;                                   // confidence level
  r.chisq = phi;                                     // value of chi-square
  r.alpha = alpha;                      // probability of larger chi-square
  r.x    = x;                        // the x-vector---for making the graph
}
